/* $Revision$ */
/* $Date$ */
/* This file (biddy-example-pp-checkoperations.c) is a C file to be included into biddy-example-pp.c */
/* Author: Robert Meolic (robert@meolic.com) */
/* This file has been released into the public domain by the author. */

/* this example is about the production planning problem */

void generateSystem () {}

void
checkOperations()
{
  char varname[2];
  Biddy_Variable v;
  Biddy_Edge tmp,support,cube,F,G,result;
  const unsigned int SIZE = 4;

  /* BASH COMMAND TO TEST IT */
  /* for i in {1..10..1}; do echo""; ./biddy-example-pp; done */

  /* for OBDDs, support is the most easiest calculated using AddVariable and AND */
  if ((Biddy_GetManagerType() == BIDDYTYPEOBDD) || (Biddy_GetManagerType() == BIDDYTYPEOBDDC)) {
    support = Biddy_GetConstantOne();
    cube = Biddy_GetConstantOne();
    for (v=0; v<SIZE; v++) {
      tmp = Biddy_AddVariableEdge();
      varname[0] = 'a' + v;
      Biddy_ChangeVariableName(Biddy_VariableTableNum()-1,varname);
      support = Biddy_And(support,tmp);
      if (v == 0) cube = Biddy_And(cube,tmp); /* cube is a subset of support */
    }
  }

  /* for ZBDDs, support is the most easiest calculated using AddElement and CHANGE */
  if ((Biddy_GetManagerType() == BIDDYTYPEZBDD) || (Biddy_GetManagerType() == BIDDYTYPEZBDDC)) {
    support = Biddy_GetBaseSet();
    cube = Biddy_GetBaseSet();
    for (v=0; v<SIZE; v++) {
      tmp = Biddy_GetElementEdge(Biddy_AddElement());
      varname[0] = 'a' + (SIZE-v-1); /* because for ZBDDs variables are added in reverse ordering */
      Biddy_ChangeVariableName(Biddy_VariableTableNum()-1,varname);
      support = Biddy_Change(support,Biddy_GetTopVariable(tmp));
      if (v == 3) cube = Biddy_Change(cube,Biddy_GetTopVariable(tmp)); /* cube is a subset of support */
    }
  }

  /* for TZBDDs, the approaches from OBDDs and ZBDDs are both working, using the first one */
  if ((Biddy_GetManagerType() == BIDDYTYPETZBDD) || (Biddy_GetManagerType() == BIDDYTYPETZBDDC)) {
    support = Biddy_GetConstantOne();
    cube = Biddy_GetConstantOne();
    for (v=0; v<SIZE; v++) {
      tmp = Biddy_AddVariableEdge();
      varname[0] = 'a' + (SIZE-v-1); /* because for TZBDDs variables are added in reverse ordering */
      Biddy_ChangeVariableName(Biddy_VariableTableNum()-1,varname);
      support = Biddy_And(support,tmp);
      if (v == 0) cube = Biddy_And(cube,tmp); /* cube is a subset of support */
    }
  }

  /* GRAPHVIZ/DOT OUTPUT OF support - DEBUGGING, ONLY */
  /*
  if (SIZE < 10) {
    Biddy_WriteDot("support.dot",support,"support",-1,FALSE);
    printf("USE 'dot -y -Tpng -O support.dot' to visualize function support.\n");
  }
  */

  /* GENERATE TWO RANDOM COMBINATION SETS */

  /* printf("Generating random combination set F...\n"); */
  /**/ F = Biddy_RandomSet(support,0.5); /**/
  /* F = Biddy_Eval2("a*b*c*!d*!e*!f*!g*!h*i*!j + !a*b*c*!d*!e*!f*!g*h*!i*!j + !a*b*c*!d*!e*!f*!g*!h*!i*!j  + a*!b*!c*d*!e*!f*!g*!h*!i*!j + !a*!b*!c*!d*e*!f*!g*!h*!i*!j + !a*!b*!c*!d*!e*!f*!g*!h*!i*j"); */
  /* F = Biddy_Eval2("!a*b*c*!d*!e*!f*!g*!h*!i*!j + !a*!b*!c*!d*e*!f*!g*!h*!i*!j + !a*!b*!c*!d*!e*!f*!g*!h*!i*j"); */
  /* F = Biddy_Eval2("!a*!b*!c + !a*b*!c + a*!b*!c + a*b*!c"); */

  /* printf("Generating random combination set G...\n"); */
  /**/ G = Biddy_RandomSet(support,0.2); /**/
  /* G = Biddy_Eval2("a*!b*!c*!d*!e*f*!g*!h*!i*j + !a*!b*c*!d*!e*!f*g*!h*!i*!j + !a*b*!c*!d*!e*!f*!g*h*i*!j"); */
  /* G = Biddy_Eval2("!a*!b*c*!d*!e*!f*g*!h*!i*!j + !a*b*!c*!d*!e*!f*!g*h*i*!j"); */
  /* G = Biddy_Eval2("!a*!b*!c + !a*b*!c + !a*!b*c + !a*b*c"); */

  result = Biddy_GetConstantZero();

  printf("Combination set for F contains %.0f combinations\n",Biddy_CountMinterms(F,0));
  Biddy_PrintfMinterms(F,FALSE);
  printf("Combination set for G contains %.0f combinations\n",Biddy_CountMinterms(G,0));
  Biddy_PrintfMinterms(G,FALSE);

  /*
  printf("\nSOME STATISTICS ABOUT F AND G\n");

  printf("Function F has %.0f minterms/combinations.\n",Biddy_CountMinterms(F,SIZE));
  printf("Function F depends on %u variables.\n",Biddy_DependentVariableNumber(F,FALSE));
  printf("%s for function F result has %u nodes (including terminals).\n",Biddy_GetManagerName(),Biddy_CountNodes(F));
  printf("%s for function F result has %u plain nodes (including terminals).\n",Biddy_GetManagerName(),Biddy_CountNodesPlain(F));
  printf("%s for function F result has %llu one-paths.\n",Biddy_GetManagerName(),Biddy_CountPaths(F));

  printf("Function G has %.0f minterms/combinations.\n",Biddy_CountMinterms(G,SIZE));
  printf("Function G depends on %u variables.\n",Biddy_DependentVariableNumber(G,FALSE));
  printf("%s for function G result has %u nodes (including terminals).\n",Biddy_GetManagerName(),Biddy_CountNodes(G));
  printf("%s for function G result has %u plain nodes (including terminals).\n",Biddy_GetManagerName(),Biddy_CountNodesPlain(G));
  printf("%s for function G result has %llu one-paths.\n",Biddy_GetManagerName(),Biddy_CountPaths(G));
  */

  /**/
  printf("ELEMENTABSTRACT (b) F\n");
  result = Biddy_ElementAbstract(F,Biddy_GetVariable("b"));
  Biddy_PrintfMinterms(result,FALSE);
  printf("ELEMENTABSTRACT (d) G\n");
  result = Biddy_ElementAbstract(G,Biddy_GetVariable("d"));
  Biddy_PrintfMinterms(result,FALSE);
  /**/

  /*
  printf("PRODUCT FxG\n");
  result = Biddy_Product(F,G);
  Biddy_PrintfMinterms(result,FALSE);
  */

  /*
  cube = Biddy_Eval2("a*b*c*d*e*!f*!g*h*!i*!j");
  printf("SELECTIVE PRODUCT FxG FOR ");
  Biddy_PrintfMinterms(cube,FALSE);
  result = Biddy_SelectiveProduct(F,G,cube);
  Biddy_PrintfMinterms(result,FALSE);
  result = Biddy_SelectiveProduct(result,G,cube);
  Biddy_PrintfMinterms(result,FALSE);
  result = Biddy_SelectiveProduct(result,G,cube);
  Biddy_PrintfMinterms(result,FALSE);
  tmp= Biddy_Diff(F,Biddy_Subset(F,result));
  Biddy_PrintfMinterms(tmp,FALSE);
  */

  /*
  printf("EXTRACT MINTERM OF F\n");
  Biddy_PrintfMinterms(Biddy_ExtractMinterm(NULL,F),FALSE);
  printf("EXTRACT MINTERM OF G\n");
  Biddy_PrintfMinterms(Biddy_ExtractMinterm(NULL,G),FALSE);
  printf("EXTRACT MINTERM OF result\n");
  Biddy_PrintfMinterms(Biddy_ExtractMinterm(NULL,result),FALSE);
  */

  /*
  printf("RESTRICTION (SUPSET) F\n");
  result = Biddy_Supset(F,G);
  Biddy_PrintfMinterms(result,FALSE);
  */

  /*
  printf("PERMISSION (SUBSET) F\n");
  result = Biddy_Subset(F,G);
  Biddy_PrintfMinterms(result,FALSE);
  */

  /*
  printf("PERMITSYM F,0\n");
  result = Biddy_Permitsym(F,0);
  Biddy_PrintfMinterms(result,FALSE);
  printf("PERMITSYM F,1\n");
  result = Biddy_Permitsym(F,1);
  Biddy_PrintfMinterms(result,FALSE);
  printf("PERMITSYM F,2\n");
  result = Biddy_Permitsym(F,2);
  Biddy_PrintfMinterms(result,FALSE);
  printf("PERMITSYM F,3\n");
  result = Biddy_Permitsym(F,3);
  Biddy_PrintfMinterms(result,FALSE);
  printf("PERMITSYM F,4\n");
  result = Biddy_Permitsym(F,4);
  Biddy_PrintfMinterms(result,FALSE);
  */

  /* GRAPHVIZ/DOT OUTPUT - DEBUGGING, ONLY */
  /*
  if (SIZE < 12) {
    Biddy_WriteDot("F.dot",F,"F",-1,FALSE);
    printf("USE 'dot -y -Tpng -O F.dot' to visualize function F.\n");
    Biddy_WriteDot("G.dot",G,"G",-1,FALSE);
    printf("USE 'dot -y -Tpng -O G.dot' to visualize function G.\n");
    Biddy_WriteDot("result.dot",result,"result",-1,FALSE);
    printf("USE 'dot -y -Tpng -O result.dot' to visualize function result.\n");
  }
  */

  /* TRUTH TABLE - DEBUGGING, ONLY */
  /*
  if (SIZE < 10) {
    printf("HERE IS A TRUTH TABLE FOR F:\n");
    Biddy_PrintfTable(F);
    printf("HERE IS A TRUTH TABLE FOR G:\n");
    Biddy_PrintfTable(G);
    printf("HERE IS A TRUTH TABLE FOR result:\n");
    Biddy_PrintfTable(result);
  }
  */

  /*
  printf("\nSOME STATISTICS ABOUT THE RESULT\n");

  printf("Function result has %.0f minterms/combinations.\n",Biddy_CountMinterms(result,SIZE));
  printf("Function represented by %s depends on %u variables.\n", Biddy_GetManagerName(), Biddy_DependentVariableNumber(result, FALSE));
  printf("%s for function result has %u nodes (including terminals).\n",Biddy_GetManagerName(),Biddy_CountNodes(result));
  printf("%s for function result has %u plain nodes (including terminals).\n",Biddy_GetManagerName(),Biddy_CountNodesPlain(result));
  printf("%s for function result has %llu one-paths.\n",Biddy_GetManagerName(),Biddy_CountPaths(result));
  */
}
