/* Linux + gcc: compile with gcc -DUNIX -o program.exe source.c -I. -L. -lbiddy -lgmp */
/* WINDOWS + MINGW: compile with gcc -DMINGW -o program.exe source.c -I. -L. -lbiddy -lgmp */
/* in this example, x1, x2, and x3 are Boolean variables */
#include "biddy.h"
int main() {
  Biddy_Edge f,g,h,r;
  Biddy_Init(); /* use default, i.e. PLAIN ROBDDS WITHOUT COMPLEMENTED EDGES */
  printf("Biddy is using %s.\n",Biddy_GetManagerName());
  f = Biddy_Eval1((Biddy_String)"(OR (AND (NOT x1) x3) (AND x1 (NOT x2) (NOT x3)))"); /* PREFIX */
  g = Biddy_Eval2((Biddy_String)"~x1*(~x3*0+x3*1)+x1*(~x2*(~x3*1+x3*0)+x2*0)"); /* INFIX */
  if (f == g) printf("Boolean functions f and g are equal.\n");
  if (f != Biddy_Not(g)) printf("Boolean functions f and ~g are not equal.\n");
  printf("Boolean function f depends on %u variables.\n",Biddy_DependentVariableNumber(f,FALSE));
  printf("Boolean function f has %.0f minterms.\n",Biddy_CountMinterms(f,0));
  printf("BDD for function f has %u nodes (including both terminals).\n",Biddy_CountNodes(f));
  printf("Here is a truth table for Boolean function f\n");
  Biddy_PrintfTable(f);
  Biddy_WriteDot("biddyexample1.dot",f,"f",-1,FALSE);
  printf("USE 'dot -y -Tpng -O biddyexample1' to visualize BDD for Boolean function f.\n");
  Biddy_Exit();
}
