/* Linux + gcc: compile with gcc -DUNIX -o program.exe source.c -I. -L. -lbiddy -lgmp */
/* WINDOWS + MINGW: compile with gcc -DMINGW -o program.exe source.c -I. -L. -lbiddy -lgmp */
/* in this example, letters D, E, H, L, O, R, and W are Boolean variables */
#include "biddy.h"
#define Str2Var(mng,x) (Biddy_Managed_GetVariable(mng,(Biddy_String)x))
int main() {
  Biddy_Manager mngobdd, mngobddc;
  Biddy_Edge f,g,r1,r2;

  Biddy_InitMNG(&mngobdd,BIDDYTYPEOBDD); /* PLAIN ROBDDS WITHOUT COMPLEMENTED EDGES */
  Biddy_InitMNG(&mngobddc,BIDDYTYPEOBDDC); /* ROBDDS WITH COMPLEMENTED EDGES */
  f = Biddy_Managed_Eval1(mngobdd,(Biddy_String)"(EXOR H E L L O (AND W O R L D))"); /* PREFIX */
  g = Biddy_Managed_Eval2(mngobdd,(Biddy_String)"~(H * E + R * O)"); /* INFIX */
  /* BASIC OPERATION */
  r1 = Biddy_Managed_Nor(mngobdd,f,g);
  /* SOP OUTPUT */
  printf("Here is Boolean function r1\n");
  Biddy_Managed_PrintfSOP(mngobdd,r1);
  /* REPLACE SOME VARIABLES: H-->L, R-->L */
  Biddy_Managed_ResetVariablesValue(mngobdd);
  Biddy_Managed_SetVariableValue(mngobdd,Str2Var(mngobdd,"H"),Biddy_Managed_GetVariableEdge(mngobdd,Str2Var(mngobdd,"L")));
  Biddy_Managed_SetVariableValue(mngobdd,Str2Var(mngobdd,"R"),Biddy_Managed_GetVariableEdge(mngobdd,Str2Var(mngobdd,"L")));
  r1 = Biddy_Managed_Replace(mngobdd,r1);
  /* SOP OUTPUT AFTER REPLACING VARIABLES */
  printf("Here is Boolean function r1 after replacing H-->L and R-->L\n");
  Biddy_Managed_PrintfSOP(mngobdd,r1);
  /* SIMPLE RESTRICTION */
  r1 = Biddy_Managed_Restrict(mngobdd,r1,Str2Var(mngobdd,"E"),FALSE);
  /* SOP OUTPUT AFTER REPLACING VARIABLES */
  printf("Here is Boolean function r1 after restricting E\n");
  Biddy_Managed_PrintfSOP(mngobdd,r1);
  /* CONVERT OBDD TO OBDDC */
  r2 = Biddy_Managed_Copy(mngobdd,mngobddc,r1);
  /* SOME STATS */
  printf("OBDD for Boolean function r1 has %u nodes.\n",Biddy_Managed_CountNodes(mngobdd,r1));
  printf("OBDDC for the same Boolean function has %u nodes.\n",Biddy_Managed_CountNodes(mngobddc,r2));
  /* GRAPHVIZ/DOT OUTPUT */
  Biddy_Managed_WriteDot(mngobdd,"biddyexample2obdd.dot",r1,"r1",-1,FALSE);
  printf("USE 'dot -y -Tpng -O biddyexample2obdd.dot' to visualize OBDD for function r1.\n");
  Biddy_Managed_WriteDot(mngobddc,"biddyexample2obddc.dot",r2,"r2",-1,FALSE);
  printf("USE 'dot -y -Tpng -O biddyexample2obddc.dot' to visualize OBDDC for function r2.\n");
  Biddy_ExitMNG(&mngobdd);
  Biddy_ExitMNG(&mngobddc);
}
